/* openssl_server.c
 * written by John Viega, Matt Messier, Pravir Chandra,
 * Eric Rescorla, Univ of Texas, and David Wilkinson
 * 09/14/03
 */
#include <unistd.h>
#include <sys/types.h>
#include <named/openssl_server.h>
#include <dns/openssl_common.h>

#define ACCESS_FILE		"/var/named/client_acl"
#define MAX_ACCESS_FILE_SIZE    1000

static int s_server_session_id_context = 1;
static DH *dh512 = NULL;
static DH *dh1024 = NULL;

typedef struct clientDN {
        char              data[256];  	/* distinguished name (DN) */
        int               size;  	/* sizeof DN */
} clientDN_t;

typedef struct access_list {
        int              clientListSize;                /* number of authorized DNs */
        clientDN_t       client[MAX_ACCESS_FILE_SIZE];  /* structure containing DN size, DN string,
                                                         * and access level */
} access_list_t;

static 
void initializeAccessList(access_list_t clientAccessList);
static 
access_list_t load_access_list(void);
static isc_result_t
checkClientCert(void *ssl_v, access_list_t clientAccessList);

void generate_eph_rsa_key(SSL_CTX *ctx) {
        RSA *rsa;
        rsa = RSA_generate_key(512, RSA_F4, NULL, NULL);
        if (!SSL_CTX_set_tmp_rsa(ctx, rsa))
                berr_exit("Couldn't set RSA key");

	RSA_free(rsa);
}

void init_dhparams(void) {
	BIO *bio;

        bio = BIO_new_file("/usr/local/ssl/dh512.pem", "r");
        if (!bio)
                berr_exit("Error opening file dh512.pem");
	dh512 = PEM_read_bio_DHparams(bio, NULL, NULL, NULL);
        if (!dh512)
                berr_exit("Error reading DH parameters from dh512.pem");
	BIO_free(bio);

	bio = BIO_new_file("/usr/local/ssl/dh1024.pem", "r");
        if (!bio)
                berr_exit("Error opening file dh1024.pem");
        dh1024 = PEM_read_bio_DHparams(bio, NULL, NULL, NULL);
        if (!dh1024)
                berr_exit("Error reading DH parameters from dh1024.pem");
        BIO_free(bio);
}

DH *tmp_dh_callback(SSL *ssl, int is_export, int keylength) {
        DH *ret;

	if (!dh512 || !dh1024)
                init_dhparams();

        switch (keylength)
        {
                case 512:
                        ret = dh512;
                        break;
                case 1024:
                default:
                        ret = dh1024;
                        break;
        }
        return ret;
}

SSL_CTX *initialize_server_ctx(void) {
        SSL_CTX *ctx;

	ctx = initialize_ctx(KEYFILE_SERVER, PASSWORD_SERVER);
        SSL_CTX_set_options(ctx, SSL_OP_ALL | SSL_OP_NO_SSLv2 |
                                 SSL_OP_SINGLE_DH_USE);
        SSL_CTX_set_tmp_dh_callback(ctx, tmp_dh_callback);
        if (SSL_CTX_set_cipher_list(ctx, CIPHER_LIST_SERVER) != 1)
                berr_exit("Error setting cipher list (no valid ciphers)");

	return ctx;
}

static void initializeAccessList(access_list_t clientAccessList) {

	clientAccessList.clientListSize = 0;
}

static access_list_t load_access_list(void) {
	access_list_t clientAccessList;
        isc_result_t result;
	FILE *f = NULL;
	char DN[256];
	int i = 0;

	/* Initialize clientAccessList */
	initializeAccessList(clientAccessList);
	
	/* Open input file */
        result = isc_stdio_open(ACCESS_FILE, "r", &f);
	if (result != ISC_R_SUCCESS) {
		printf("Error opening access file\n");
		return clientAccessList;
	}

	/* Load file into clientAccessList structure */	
	while (i < MAX_ACCESS_FILE_SIZE && fgets(DN, 256, f) != NULL) {
		strncpy(clientAccessList.client[i].data, DN, strlen(DN) - 1);
		clientAccessList.client[i].size = strlen(DN) - 1;
		i++;
	}
	
	if (i == MAX_ACCESS_FILE_SIZE)
		printf("Warning: File size may be larger than maximum allowed\n");
	
	clientAccessList.clientListSize = i;

	/* Close input file */
        result = isc_stdio_close(f);
        if (result != ISC_R_SUCCESS) {
		printf("Error closing access file\n");
		clientAccessList.clientListSize = 0;
		return clientAccessList;
	}
	
	return clientAccessList;	
}
	
static isc_result_t
checkClientCert(void *ssl_v, access_list_t clientAccessList) {
	SSL *ssl = (SSL *)ssl_v;
	X509 *peerCert;
	char peerName[256];
	int i;
	int peerNameSize = 0;

        for (i = 0; i < 256; i++)
		peerName[i] = '\0';

	/* Get client certificate */	
	if ((peerCert = SSL_get_peer_certificate(ssl)) == NULL) {
		fprintf(stderr, "Error: no peer cert\n");
		return ISC_R_FAILURE;
	}

	/* Extract the client's Distinguished Name */
	X509_NAME_oneline(X509_get_subject_name(peerCert),
			  peerName, sizeof(peerName));
	peerNameSize = strlen(peerName);

	/* Search clientAccessList for a match */
	for (i = 0; i < clientAccessList.clientListSize; i++) {
		if (clientAccessList.client[i].size != peerNameSize)
			continue;
		if (strncmp(clientAccessList.client[i].data, peerName, peerNameSize) == 0)
			break;
	}

	/* Return failure if no match found */
	if (i >= clientAccessList.clientListSize)
		return ISC_R_FAILURE;

	return ISC_R_SUCCESS;
}

isc_result_t ssl_authenticate_client(void) {
        SSL_CTX *ctx;
        SSL *ssl;
        BIO *sbio;
        int sock, s;
	isc_result_t ret;
        int result;
        struct sockaddr_in sin;
        int val = 1;
	access_list_t clientAccessList;
	int iteration = 0;

        /* Load list of authorized clients from file */
        clientAccessList = load_access_list();
        if (clientAccessList.clientListSize == 0)
                return ISC_R_FAILURE;

retry:	
        /* Get file descriptor and create socket */
        sock = socket(AF_INET, SOCK_STREAM, 0);
        if (sock < 0) {
		printf("pid %d: Error, couldn't make socket\n", getpid());
		return ISC_R_FAILURE;
	}

        memset(&sin, 0, sizeof(sin));
        sin.sin_addr.s_addr = INADDR_ANY;
        sin.sin_family = AF_INET;
        sin.sin_port = htons(PORT);
        setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &val, sizeof(val));

	/* Bind to socket */
        if (bind(sock, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
                printf("pid %d: Error, couldn't bind\n", getpid());
	        close(sock);
		return ISC_R_FAILURE;
	}

	/* Listen for connection */
        listen(sock, 5);

        s = accept(sock, 0, 0);
	
        if (s < 0) {
                printf("pid %d: Error, problem accepting\n", getpid());
                close(sock);
		return ISC_R_FAILURE;
	}

	RAND_cleanup();

        /* Seed random number generator */
        if (!(RAND_load_file(RANDOM_SERVER, 1024))) {
                printf("Couldn't load randomness\n");
                return ISC_R_FAILURE;
        }
	
        /* Create SSL context */
        ctx = initialize_server_ctx();

        SSL_CTX_set_session_id_context(ctx, (void *)&s_server_session_id_context,
                                       sizeof s_server_session_id_context);
	
	/* Create ssl object and accept ssl connection */
	sbio = BIO_new(BIO_s_socket());
	BIO_set_fd(sbio, s, BIO_NOCLOSE);
	ssl = SSL_new(ctx);
	SSL_set_accept_state(ssl);
        SSL_set_bio(ssl, sbio, sbio);

        result = SSL_accept(ssl);

        switch(SSL_get_error(ssl, result))
        {
                case SSL_ERROR_NONE:
       	                break;
                case SSL_ERROR_ZERO_RETURN:
                        printf("Zero return error\n");
                        break;
                case SSL_ERROR_WANT_READ:
                        printf("Want read error\n");
                        break;
                case SSL_ERROR_WANT_WRITE:
                        printf("Want write error\n");
                        break;
                default:
                        printf("An SSL_accept error occurred\n");
			iteration++;
			if (iteration < 2) {
			        SSL_shutdown(ssl);
			        SSL_free(ssl);
				SSL_CTX_free(ctx);
				close(s);
				close(sock);
				ERR_remove_state(0);
				sleep(3);
				goto retry;
			}
                        break;
        }
	
        if (result <= 0) {
                printf("pid %d: SSL_accept error, result = %d\n", getpid(), result);
		ret = ISC_R_FAILURE;
		goto cleanup;
	}
	
	/* Check to see if client is authorized to perform nsreroute */
	ret = checkClientCert(ssl, clientAccessList);
	if (ret == ISC_R_FAILURE)
		printf("Client not authorized for nsreroute\n");

cleanup:

	/* Close ssl and underlying connection, and return error code */
        SSL_shutdown(ssl);
        SSL_free(ssl);
	SSL_CTX_free(ctx);
        close(s);
        close(sock);
	ERR_remove_state(0);
	return ret;
}

