/*
 * Copyright (C) 1999-2002  Internet Software Consortium.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 */

/* $Id: reroute.c by ISC and David Wilkinson, 06-23-03 */

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <unistd.h>
#include <fcntl.h>
#include <stddef.h>

#include <sys/file.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>

#include <isc/print.h>
#include <isc/string.h>
#include <isc/taskpool.h>
#include <isc/util.h>
#include <isc/stdio.h>
#include <isc/types.h>
#include <isc/file.h>
#include <isc/string.h>
#include <isc/lang.h>
#include <isc/sockaddr.h>
#include <isc/app.h>
#include <isc/base64.h>
#include <isc/buffer.h>
#include <isc/commandline.h>
#include <isc/entropy.h>
#include <isc/event.h>
#include <isc/lex.h>
#include <isc/mem.h>
#include <isc/region.h>
#include <isc/socket.h>
#include <isc/task.h>
#include <isc/timer.h>

#include <dns/db.h>
#include <dns/dbiterator.h>
#include <dns/diff.h>
#include <dns/dnssec.h>
#include <dns/events.h>
#include <dns/fixedname.h>
#include <dns/journal.h>
#include <dns/masterdump.h>
#include <dns/message.h>
#include <dns/nxt.h>
#include <dns/rdataclass.h>
#include <dns/rdataset.h>
#include <dns/rdatastruct.h>
#include <dns/rdatasetiter.h>
#include <dns/soa.h>
#include <dns/ssu.h>
#include <dns/view.h>
#include <dns/zone.h>
#include <dns/zt.h>
#include <dns/name.h>
#include <dns/types.h>
#include <dns/openssl_common.h>

#include <named/client.h>
#include <named/log.h>
#include <named/update.h>
#include <named/reroute.h>
#include <named/server.h>
#include <named/openssl_server.h>

#ifndef NAMED_CONF
#define NAMED_CONF "/etc/named.conf"
#endif

#ifndef OUTPUTBUF
#define OUTPUTBUF 32767
#endif

/**************************************************************************/

/*
 * Log level for tracing dynamic update protocol requests.
 */
#define LOGLEVEL_PROTOCOL	ISC_LOG_INFO

/*
 * Log level for low-level debug tracing.
 */
#define LOGLEVEL_DEBUG 		ISC_LOG_DEBUG(8)

/*
 * Check an operation for failure.  These macros all assume that
 * the function using them has a 'result' variable and a 'failure'
 * label.
 */
#define CHECK(op) \
	do { result = (op); 				  	 \
	       if (result != ISC_R_SUCCESS) goto failure; 	 \
	} while (0)

/*
 * Fail unconditionally with result 'code', which must not
 * be ISC_R_SUCCESS.  The reason for failure presumably has
 * been logged already.
 *
 * The test against ISC_R_SUCCESS is there to keep the Solaris compiler
 * from complaining about "end-of-loop code not reached".
 */

#define FAIL(code) \
	do {							\
		result = (code);				\
		if (result != ISC_R_SUCCESS) goto failure;	\
	} while (0)

/*
 * Fail unconditionally and log as a client error.
 * The test against ISC_R_SUCCESS is there to keep the Solaris compiler
 * from complaining about "end-of-loop code not reached".
 */
#define FAILC(code, msg) \
	do {							\
		result = (code);				\
		update_log(client, zone, LOGLEVEL_PROTOCOL,   	\
			      "update failed: %s (%s)",		\
		      	      msg, isc_result_totext(result));	\
		if (result != ISC_R_SUCCESS) goto failure;	\
	} while (0)

/*
 * Fail unconditionally and log as a server error.
 * The test against ISC_R_SUCCESS is there to keep the Solaris compiler
 * from complaining about "end-of-loop code not reached".
 */
#define FAILS(code, msg) \
	do {							\
		result = (code);				\
		update_log(client, zone, LOGLEVEL_PROTOCOL,		\
			      "error: %s: %s", 			\
			      msg, isc_result_totext(result));	\
		if (result != ISC_R_SUCCESS) goto failure;	\
	} while (0)

/**************************************************************************/
static double t1, t2, elapsed;
struct timeval tp;
int rtn;

typedef struct reroute_event reroute_event_t;

struct reroute_event {
	ISC_EVENT_COMMON(reroute_event_t);
	dns_zone_t 		*zone;
	isc_result_t		result;
	dns_message_t		*answer;
};

/**************************************************************************/
/*
 * Forward declarations.
 */

static void reroute_action(isc_task_t *task, isc_event_t *event);
static isc_result_t append_toconfig(dns_message_t *request, isc_mem_t *mctx);
static void reroutedone_action(isc_task_t *task, isc_event_t *event);
static isc_result_t write_zone(dns_message_t *request, isc_mem_t *mctx);

/**************************************************************************/

static void
update_log(ns_client_t *client, dns_zone_t *zone,
	   int level, const char *fmt, ...) ISC_FORMAT_PRINTF(4, 5);

static void
update_log(ns_client_t *client, dns_zone_t *zone,
	   int level, const char *fmt, ...)
{
	va_list ap;
	char message[4096];
	char namebuf[DNS_NAME_FORMATSIZE];
	char classbuf[DNS_RDATACLASS_FORMATSIZE];

	if (client == NULL || zone == NULL)
		return;

	if (isc_log_wouldlog(ns_g_lctx, level) == ISC_FALSE)
		return;

	dns_name_format(dns_zone_getorigin(zone), namebuf,
			sizeof(namebuf));
	dns_rdataclass_format(dns_zone_getclass(zone), classbuf,
			      sizeof(classbuf));

	va_start(ap, fmt);
	vsnprintf(message, sizeof message, fmt, ap);
	va_end(ap);

	ns_client_log(client, NS_LOGCATEGORY_UPDATE, NS_LOGMODULE_UPDATE,
		      level, "updating zone '%s/%s': %s",
		      namebuf, classbuf, message);
}

static isc_result_t
send_reroute_event(ns_client_t *client, dns_zone_t *zone) {
	isc_result_t result = ISC_R_SUCCESS;
	reroute_event_t *event = NULL;
	isc_task_t *zonetask = NULL;
	ns_client_t *evclient;

	event = (reroute_event_t *)
		isc_event_allocate(client->mctx, client, DNS_EVENT_REROUTE,
				   reroute_action, NULL, sizeof(*event));
	if (event == NULL)
		FAIL(ISC_R_NOMEMORY);
	event->zone = zone;
	event->result = ISC_R_SUCCESS;

	evclient = NULL;
	ns_client_attach(client, &evclient);
	event->ev_arg = evclient;

	dns_zone_gettask(zone, &zonetask);
	isc_task_send(zonetask, (isc_event_t **)&event);

 failure:
	if (event != NULL)
		isc_event_free((isc_event_t **)&event);
	return (result);
}

static void
respond(ns_client_t *client, isc_result_t result) {
	isc_result_t msg_result;

	msg_result = dns_message_reply(client->message, ISC_TRUE);
	if (msg_result != ISC_R_SUCCESS)
		goto msg_failure;
	client->message->rcode = dns_result_torcode(result);

	ns_client_send(client);
	return;

 msg_failure:
	isc_log_write(ns_g_lctx, NS_LOGCATEGORY_UPDATE, NS_LOGMODULE_UPDATE,
		      ISC_LOG_ERROR,
		      "could not create update response message: %s",
		      isc_result_totext(msg_result));
	ns_client_next(client, msg_result);
}

void
ns_reroute_start(ns_client_t *client, isc_result_t sigresult) {
	dns_message_t *request = client->message;
	isc_result_t result;
	dns_name_t *zonename;
	dns_rdataset_t *zone_rdataset;
	dns_zone_t *zone = NULL;

        rtn = gettimeofday(&tp, NULL);             /* get current time */
        t1 = (double)tp.tv_sec+(1.e-6)*tp.tv_usec;

	/*
	 * Interpret the zone section.
	 */
	result = dns_message_firstname(request, DNS_SECTION_ZONE);
	if (result != ISC_R_SUCCESS)
		FAILC(DNS_R_FORMERR,
		      "reroute zone section empty");

	/*
	 * The zone section contains exactly one "question", and
	 * it must be of type SOA.
	 */
	
	zonename = NULL;
	dns_message_currentname(request, DNS_SECTION_ZONE, &zonename);
	zone_rdataset = ISC_LIST_HEAD(zonename->list);
	if (zone_rdataset->type != dns_rdatatype_soa)
		FAILC(DNS_R_FORMERR,
		      "reroute zone section contains non-SOA");
	if (ISC_LIST_NEXT(zone_rdataset, link) != NULL)
		FAILC(DNS_R_FORMERR,
		      "reroute zone section contains multiple RRs");

	/* The zone section must have exactly one name. */
	result = dns_message_nextname(request, DNS_SECTION_ZONE);
	if (result != ISC_R_NOMORE)
		FAILC(DNS_R_FORMERR,
		      "reroute zone section contains multiple RRs");

	result = dns_zt_find(client->view->zonetable, zonename, 0, NULL,
			     &zone);
	if (result != ISC_R_SUCCESS)
		FAILC(DNS_R_NOTAUTH,
		      "not authoritative for update zone");

	switch(dns_zone_gettype(zone)) {
	case dns_zone_master:
		
		/* We can now fail due to a bad signature as we now know
		 * that we are the master.
		 */

		if (sigresult != ISC_R_SUCCESS)
			FAIL(sigresult);

		CHECK(send_reroute_event(client, zone));
		break;
	default:
		FAILC(DNS_R_NOTAUTH,
		      "not authoritative for reroute zone");
	}
	return;

 failure:
	/*
	 * We failed without having sent an update event to the zone.
	 * We are still in the client task context, so we can
	 * simply give an error response without switching tasks.
	 */
	respond(client, result);
	if (zone != NULL)
		dns_zone_detach(&zone);
}

static void
reroute_action(isc_task_t *task, isc_event_t *event) {
	reroute_event_t *uev = (reroute_event_t *) event;
	ns_client_t *client = (ns_client_t *)event->ev_arg;
	isc_result_t result;
	isc_mem_t *mctx = client->mctx;
	dns_message_t *request = client->message;
	
	INSIST(event->ev_type == DNS_EVENT_REROUTE);

	/* Authenticate client */
	result = ssl_authenticate_client();
       	if (result != ISC_R_SUCCESS)
		goto failure;

	/* Create new zone file */
        result = write_zone(request, mctx);
        if (result != ISC_R_SUCCESS)
		goto failure;
						
	/* Create new zone entry in /etc/named.conf */
	result = append_toconfig(request, mctx);
        if (result != ISC_R_SUCCESS)
		goto failure;

 failure:
	isc_task_detach(&task);
	uev->result = result;
	uev->ev_type = DNS_EVENT_REROUTEDONE;
	uev->ev_action = reroutedone_action;
	isc_task_send(client->task, &event);
	INSIST(event == NULL);
}

static isc_result_t
append_toconfig(dns_message_t *msg, isc_mem_t *mctx) {
	isc_result_t result;
	dns_name_t *victim_dns = NULL;
        dns_name_t *prefix = NULL;
        dns_name_t *suffix = NULL;
	isc_buffer_t *suffix_buf = NULL;
	int fd;
	unsigned int len = OUTPUTBUF;
	unsigned int labels = 0;
	char victim_domain_ptr[DNS_NAME_FORMATSIZE];
	
	/* Get first name of Reroute (or Update) section of msg */
	result = dns_message_firstname(msg, DNS_SECTION_REROUTE);
	if (result != ISC_R_SUCCESS)
		return result;
	dns_message_currentname(msg, DNS_SECTION_REROUTE, &victim_dns);

	result = dns_message_gettempname(msg, &suffix);
	if (result != ISC_R_SUCCESS)
		return result;

	result = isc_buffer_allocate(mctx, &suffix_buf, len);
	if (result != ISC_R_SUCCESS) {
		dns_message_puttempname(msg, &suffix);
		return result;
	}
	
	/* Store domain of victim_dns in suffix */
	dns_name_setbuffer(suffix, suffix_buf);
	labels = dns_name_countlabels(victim_dns);
	result = dns_name_splitatdepth(victim_dns, labels-1, prefix, suffix);
	if (result != ISC_R_SUCCESS) {
		dns_message_puttempname(msg, &suffix);
		return result;
	}

	/* Point victim_domain_ptr to domain name of victim DNS server */
	dns_name_format(suffix, victim_domain_ptr, DNS_NAME_FORMATSIZE);

        /* Append entry to named.conf */
	fd = open(NAMED_CONF, O_APPEND|O_RDWR);
	if (fd == -1) {
		dns_message_puttempname(msg, &suffix);
                return ISC_R_IOERROR;	/* error opening file */
	}
	if (flock(fd, LOCK_EX) == -1) {
		dns_message_puttempname(msg, &suffix);
		return ISC_R_IOERROR;	/* couldn't get exclusive lock on file */
	}
	
	/* Write contents of zone statement */
	write(fd, "\n", 1);
	write(fd, "zone \"", 6);
	write(fd, victim_domain_ptr, strlen(victim_domain_ptr));
	write(fd, "\" IN {\n", 7);
	write(fd, "\ttype master;\n", 14);
	write(fd, "\tfile \"db.", 10);
	write(fd, victim_domain_ptr, strlen(victim_domain_ptr));
	write(fd, "\";\n", 3);
	write(fd, "};\n", 3);	

	if (flock(fd, LOCK_UN) == -1) {
		dns_message_puttempname(msg, &suffix);
		return ISC_R_IOERROR;	/* couldn't unlock file */
	}
        if (close(fd) == -1) {
		dns_message_puttempname(msg, &suffix);
                return ISC_R_IOERROR;	/* error closing file */
	}

	/* Clean up */	
	isc_buffer_free(&suffix_buf);
        dns_message_puttempname(msg, &suffix);
	return ISC_R_SUCCESS;
}

static isc_result_t 
write_zone(dns_message_t *msg, isc_mem_t *mctx) {
        isc_result_t result;
        dns_name_t *victim_name = NULL;
	dns_name_t *victim_name2 = NULL;
        dns_master_style_t *style;
	dns_rdataset_t *rdataset;
	dns_rdataset_t *rdataset2;
	dns_rdataset_t *rdataset_proxy;
	dns_rdataset_t *rdataset_proxy2;
	dns_name_t *server_domain_name = NULL;
	dns_name_t *victim_suffix = NULL;
	isc_buffer_t *victim_suffix_buf = NULL;
	isc_buffer_t *victim_dns_buf = NULL;
	isc_buffer_t *victim_dns_buf2 = NULL;
        isc_buffer_t *buf = NULL;
	isc_buffer_t *buf2 = NULL;
	isc_buffer_t *proxy_buf = NULL;
	isc_buffer_t *proxy_buf2 = NULL;
        unsigned int len = OUTPUTBUF;
	unsigned int labels = 0;
        char victim_domain_ptr[DNS_NAME_FORMATSIZE];
	char server_domain_ptr[DNS_NAME_FORMATSIZE];
        char filename_ptr[DNS_NAME_FORMATSIZE];
	char host_name[256];
	int fd;

	style = &dns_master_style_default;
       
        result = dns_message_firstname(msg, DNS_SECTION_REROUTE);
        if (result != ISC_R_SUCCESS)
                return result;

	/* Put the first victim DNS server name and its IP address into buf */
	dns_message_currentname(msg, DNS_SECTION_REROUTE, &victim_name);
        rdataset = victim_name->list.head;
        result = isc_buffer_allocate(mctx, &buf, len);
	if (result != ISC_R_SUCCESS)
                return result;
        result = dns_master_rdatasettotext(victim_name, rdataset, style, buf);
        if (result != ISC_R_SUCCESS)
                return result;
	
        /* Put the second victim DNS server name and its IP address into buf2 */
	result = dns_message_nextname(msg, DNS_SECTION_REROUTE);
        dns_message_currentname(msg, DNS_SECTION_REROUTE, &victim_name2);
        rdataset2 = victim_name2->list.head;
        result = isc_buffer_allocate(mctx, &buf2, len);
	if (result != ISC_R_SUCCESS)
                return result;
	result = dns_master_rdatasettotext(victim_name2, rdataset2, style, buf2);
        if (result != ISC_R_SUCCESS)
                return result;
	
	/* Put the first victim DNS server name and a list
         * of proxy server IP addresses into proxy_buf */
        rdataset_proxy = victim_name->list.head->link.next;
        result = isc_buffer_allocate(mctx, &proxy_buf, len);
	if (result != ISC_R_SUCCESS)
                return result;
	result = dns_master_rdatasettotext(victim_name, 
					   rdataset_proxy, style, proxy_buf);
        if (result != ISC_R_SUCCESS)
                return result;

        /* Put the second victim DNS server name and a list
         * of proxy server IP addresses into proxy_buf2 */
        rdataset_proxy2 = victim_name2->list.head->link.next;
        result = isc_buffer_allocate(mctx, &proxy_buf2, len);
	if (result != ISC_R_SUCCESS)
                return result;
	result = dns_master_rdatasettotext(victim_name2,
                                           rdataset_proxy2, style, proxy_buf2);
        if (result != ISC_R_SUCCESS)
                return result;
	
	/* Get domain name of victim.  Put into victim_suffix. */
	result = dns_message_gettempname(msg, &victim_suffix);
        if (result != ISC_R_SUCCESS)
                return result;
	
	result = isc_buffer_allocate(mctx, &victim_suffix_buf, len);
	if (result != ISC_R_SUCCESS) {
		dns_message_puttempname(msg, &victim_suffix);
		return result;
	}
	
        dns_name_setbuffer(victim_suffix, victim_suffix_buf);
        labels = dns_name_countlabels(victim_name);
        result = dns_name_splitatdepth(victim_name, labels-1, NULL, victim_suffix);
        if (result != ISC_R_SUCCESS) {
		dns_message_puttempname(msg, &victim_suffix);
                return result;
	}
	
	/* Point victim_domain_ptr to domain name of victim DNS server */
        dns_name_format(victim_suffix, victim_domain_ptr, DNS_NAME_FORMATSIZE);

     	/* This host is the server of authority */
        result = gethostname(host_name, 255);     /* get FQDN of current host */

	/* Put the FQDN of first victim's DNS server into victim_dns_buf */
        result = isc_buffer_allocate(mctx, &victim_dns_buf, len);
        result = dns_name_tofilenametext(victim_name, ISC_FALSE, victim_dns_buf);

        /* Put the FQDN of second victim's DNS server into victim_dns_buf2 */
        result = isc_buffer_allocate(mctx, &victim_dns_buf2, len);
        result = dns_name_tofilenametext(victim_name2, ISC_FALSE, victim_dns_buf2);
	
	/* Point server_domain_ptr to domain name of SOA server */	
        result = dns_message_firstname(msg, DNS_SECTION_QUESTION);
        if (result != ISC_R_SUCCESS) {
		dns_message_puttempname(msg, &victim_suffix);
                return result;
	}
	
        dns_message_currentname(msg, DNS_SECTION_QUESTION, &server_domain_name);
        dns_name_format(server_domain_name, server_domain_ptr, DNS_NAME_FORMATSIZE);

	/* Construct name of zone file using "db." followed by domain name of victim */	
        strcpy(filename_ptr, ZONE_DIR);   /* ZONE_DIR is where zone files reside */
	strcat(filename_ptr, "/db.");
        strncat(filename_ptr, victim_domain_ptr, strlen(victim_domain_ptr));

	/* Open new file with exclusive write access */
        fd = open(filename_ptr, O_CREAT|O_WRONLY|O_EXCL, 
		  S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
        if (fd == -1) {
		printf("Could not create new zone file\n");
		isc_buffer_free(&buf);
               	isc_buffer_free(&proxy_buf);
	        isc_buffer_free(&victim_suffix_buf);
                isc_buffer_free(&victim_dns_buf);
		dns_message_puttempname(msg, &victim_suffix);
		return ISC_R_IOERROR;   	/* file already exists or other error */
	}

	/* Write contents of new zone file */	
	write(fd, "@ IN SOA ", 9);
	write(fd, host_name, strlen(host_name));
	write(fd, ". root.", 7);
	write(fd, server_domain_ptr, strlen(server_domain_ptr));
	write(fd, ". (\n", 4);
	write(fd, "\t\t\t\t1\t; Serial\n", 15);
	write(fd, "\t\t\t\t3h\t; Refresh after 3 hours\n", 31);
	write(fd, "\t\t\t\t1h\t; Retry after 1 hour\n", 28);
	write(fd, "\t\t\t\t1w\t; Expire after 1 week\n", 29);
	write(fd, "\t\t\t\t1h\t; Negative caching TTL of 1 hour\n", 40);
	write(fd, "\t\t\t\t)\n\n", 7);
	write(fd, "\t\t\tIN NS\t", 9);
        write(fd, (char *)isc_buffer_base(victim_dns_buf),
              (int)isc_buffer_usedlength(victim_dns_buf));
	write(fd, "\n", 1);
	write(fd, "\t\t\tIN NS\t", 9);
	write(fd, (char *)isc_buffer_base(victim_dns_buf2),
              (int)isc_buffer_usedlength(victim_dns_buf2));	
	write(fd, "\n\n", 2);
	write(fd, (char *)isc_buffer_base(buf),
              (int)isc_buffer_usedlength(buf));
	write(fd, (char *)isc_buffer_base(proxy_buf),
              (int)isc_buffer_usedlength(proxy_buf));
	write(fd, "\n\n", 1);
	write(fd, (char *)isc_buffer_base(buf2),
              (int)isc_buffer_usedlength(buf2));
        write(fd, (char *)isc_buffer_base(proxy_buf2),
              (int)isc_buffer_usedlength(proxy_buf2));
	write(fd, "\n", 1);
	
        if (close(fd) == -1) {
		dns_message_puttempname(msg, &victim_suffix);
		return ISC_R_IOERROR;   /* error closing file */
	}

	/* Clean up */	
	isc_buffer_free(&buf);
	isc_buffer_free(&buf2);
	isc_buffer_free(&proxy_buf);
	isc_buffer_free(&proxy_buf2);
	isc_buffer_free(&victim_suffix_buf);
	isc_buffer_free(&victim_dns_buf);
	isc_buffer_free(&victim_dns_buf2);
        dns_message_puttempname(msg, &victim_suffix);
        return ISC_R_SUCCESS;
}

static void
reroutedone_action(isc_task_t *task, isc_event_t *event) {
	reroute_event_t *uev = (reroute_event_t *) event;
	ns_client_t *client = (ns_client_t *) event->ev_arg;
	isc_boolean_t new_zone = ISC_FALSE;
	
	UNUSED(task);

	INSIST(event->ev_type == DNS_EVENT_REROUTEDONE);
	INSIST(task == client->task);

	if (uev->result == ISC_R_SUCCESS)
		new_zone = ISC_TRUE;

	respond(client, uev->result);
	ns_client_detach(&client);
	isc_event_free(&event);
	printf("pid %d, parent pid %d\n", getpid(), getppid());
	
	/* Reload zones all zones into memory if a new one has been added */
	if (new_zone)
		ns_server_reloadwanted(ns_g_server);

        rtn = gettimeofday(&tp, NULL);             /* get current time */
        t2 = (double)tp.tv_sec+(1.e-6)*tp.tv_usec;

        elapsed = t2 - t1;
        printf("server processing time is %g seconds\n", elapsed);
}

