/*
 * Copyright (C) 2000-2002  Internet Software Consortium.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 */

/* $Id: nsreroute.c (derived from nsupdate.c by ISC) by David Wilkinson, 06-24-03 */

#include "../../config.h"

#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <netdb.h>
#include <stdlib.h>
#include <unistd.h>
#include <stddef.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>

#include <isc/app.h>
#include <isc/base64.h>
#include <isc/buffer.h>
#include <isc/commandline.h>
#include <isc/entropy.h>
#include <isc/event.h>
#include <isc/lex.h>
#include <isc/mem.h>
#include <isc/region.h>
#include <isc/sockaddr.h>
#include <isc/socket.h>
#include <isc/stdio.h>
#include <isc/string.h>
#include <isc/task.h>
#include <isc/timer.h>
#include <isc/types.h>
#include <isc/util.h>

#include <dns/callbacks.h>
#include <dns/dispatch.h>
#include <dns/events.h>
#include <dns/fixedname.h>
#include <dns/masterdump.h>
#include <dns/message.h>
#include <dns/name.h>
#include <dns/rdata.h>
#include <dns/rdataclass.h>
#include <dns/rdatalist.h>
#include <dns/rdataset.h>
#include <dns/rdatastruct.h>
#include <dns/rdatatype.h>
#include <dns/request.h>
#include <dns/result.h>
#include <dns/tsig.h>
#include <dns/types.h>
#include <dns/openssl_common.h>

#include <dst/dst.h>

#include <lwres/lwres.h>
#include <lwres/net.h>

#ifdef HAVE_ADDRINFO
#ifdef HAVE_GETADDRINFO
#ifdef HAVE_GAISTRERROR
#define USE_GETADDRINFO
#endif
#endif
#endif

#ifndef USE_GETADDRINFO
#ifndef ISC_PLATFORM_NONSTDHERRNO
extern int h_errno;
#endif
#endif

#define MAXCMD (4 * 1024)
#define INITDATA (32 * 1024)
#define MAXDATA (64 * 1024)
#define NAMEBUF 512
#define WORDLEN 512
#define PACKETSIZE ((64 * 1024) - 1)
#define INITTEXT (2 * 1024)
#define MAXTEXT (128 * 1024)
#define FIND_TIMEOUT 100000
#define TTL_MAX 2147483647	/* Maximum signed 32 bit integer. */

#define DNSDEFAULTPORT 53

#ifndef RESOLV_CONF
#define RESOLV_CONF "/etc/resolv.conf"
#endif

#ifndef OUTPUTBUF
#define OUTPUTBUF 32767
#endif

static isc_boolean_t debugging = ISC_FALSE, ddebugging = ISC_FALSE;
static isc_boolean_t memdebugging = ISC_FALSE;
static isc_boolean_t have_ipv4 = ISC_FALSE;
static isc_boolean_t have_ipv6 = ISC_FALSE;
static isc_boolean_t is_dst_up = ISC_FALSE;
static isc_boolean_t usevc = ISC_FALSE;
static isc_taskmgr_t *taskmgr = NULL;
static isc_task_t *global_task = NULL;
static isc_event_t *global_event = NULL;
static isc_mem_t *mctx = NULL;
static dns_dispatchmgr_t *dispatchmgr = NULL;
static dns_requestmgr_t *requestmgr = NULL;
static isc_socketmgr_t *socketmgr = NULL;
static isc_timermgr_t *timermgr = NULL;
static dns_dispatch_t *dispatchv4 = NULL;
static dns_dispatch_t *dispatchv6 = NULL;
static dns_message_t *reroutemsg = NULL;
static dns_fixedname_t resolvdomain; /* from resolv.conf's domain line */
static dns_name_t *origin; /* Points to one of above, or dns_rootname */
static dns_name_t *userzone = NULL;
static dns_tsigkey_t *key = NULL;
static lwres_context_t *lwctx = NULL;
static lwres_conf_t *lwconf;
static isc_sockaddr_t *servers;
static int ns_inuse = 0;
static int ns_total = 0;
static isc_sockaddr_t *userserver = NULL;
static isc_sockaddr_t *localaddr = NULL;
static isc_entropy_t *entp = NULL;
static isc_boolean_t shuttingdown = ISC_FALSE;
static FILE *input;
static isc_boolean_t interactive = ISC_TRUE;
static isc_boolean_t seenerror = ISC_FALSE;
static const dns_master_style_t *style;
static int requests = 0;
static dns_name_t *client_hostname;    // reroute
static int starting_pid = 0;
static double t1;	/* start time */

typedef struct nsu_requestinfo {
	dns_message_t *msg;
	isc_sockaddr_t *addr;
} nsu_requestinfo_t;

static void
sendrequest(isc_sockaddr_t *srcaddr, isc_sockaddr_t *destaddr,
	    dns_message_t *msg, dns_request_t **request);
static void
fatal(const char *format, ...) ISC_FORMAT_PRINTF(1, 2);

static void
debug(const char *format, ...) ISC_FORMAT_PRINTF(1, 2);

static void
ddebug(const char *format, ...) ISC_FORMAT_PRINTF(1, 2);

#define STATUS_MORE	(isc_uint16_t)0
#define STATUS_SEND	(isc_uint16_t)1
#define STATUS_QUIT	(isc_uint16_t)2
#define STATUS_SYNTAX	(isc_uint16_t)3

static void
fatal(const char *format, ...) {
	va_list args;

	va_start(args, format);
	vfprintf(stderr, format, args);
	va_end(args);
	fprintf(stderr, "\n");
	exit(1);
}

static void
debug(const char *format, ...) {
	va_list args;

	if (debugging) {
		va_start(args, format);
		vfprintf(stderr, format, args);
		va_end(args);
		fprintf(stderr, "\n");
	}
}

static void
ddebug(const char *format, ...) {
	va_list args;

	if (ddebugging) {
		va_start(args, format);
		vfprintf(stderr, format, args);
		va_end(args);
		fprintf(stderr, "\n");
	}
}

static inline void
check_result(isc_result_t result, const char *msg) {
	if (result != ISC_R_SUCCESS)
		fatal("%s: %s", msg, isc_result_totext(result));
}

static void *
mem_alloc(void *arg, size_t size) {
	return (isc_mem_get(arg, size));
}

static void
mem_free(void *arg, void *mem, size_t size) {
	isc_mem_put(arg, mem, size);
}

static char *
nsu_strsep(char **stringp, const char *delim) {
	char *string = *stringp;
	char *s;
	const char *d;
	char sc, dc;

	if (string == NULL)
		return (NULL);

	for (; *string != '\0'; string++) {
		sc = *string;
		for (d = delim; (dc = *d) != '\0'; d++) {
			if (sc == dc)
				break;
		}
		if (dc == 0)
			break;
	}

	for (s = string; *s != '\0'; s++) {
		sc = *s;
		for (d = delim; (dc = *d) != '\0'; d++) {
			if (sc == dc) {
				*s++ = '\0';
				*stringp = s;
				return (string);
			}
		}
	}
	*stringp = NULL;
	return (string);
}

static unsigned int
count_dots(char *s, isc_boolean_t *last_was_dot) {
	int i = 0;
	*last_was_dot = ISC_FALSE;
	while (*s != 0) {
		if (*s++ == '.') {
			i++;
			*last_was_dot = ISC_TRUE;
		} else
			*last_was_dot = ISC_FALSE;
	}
	return (i);
}

static void
reset_system(void) {
	isc_result_t result;
		
	ddebug("reset_system()");
	
	/* If the reroute message is still around, destroy it */
	if (reroutemsg != NULL)
		dns_message_reset(reroutemsg, DNS_MESSAGE_INTENTRENDER);
	else {
		result = dns_message_create(mctx, DNS_MESSAGE_INTENTRENDER,
					    &reroutemsg);
		check_result(result, "dns_message_create");
	}
	reroutemsg->opcode = dns_opcode_reroute;
}

static void
doshutdown(void) {
	isc_task_detach(&global_task);

	if (userserver != NULL)
		isc_mem_put(mctx, userserver, sizeof(isc_sockaddr_t));

	if (localaddr != NULL)
		isc_mem_put(mctx, localaddr, sizeof(isc_sockaddr_t));

	if (key != NULL) {
		ddebug("Freeing key");
		dns_tsigkey_detach(&key);
	}

	if (reroutemsg != NULL)
		dns_message_destroy(&reroutemsg);

	if (is_dst_up) {
		ddebug("Destroy DST lib");
		dst_lib_destroy();
		is_dst_up = ISC_FALSE;
	}

	if (entp != NULL) {
		ddebug("Detach from entropy");
		isc_entropy_detach(&entp);
	}

	lwres_conf_clear(lwctx);
	lwres_context_destroy(&lwctx);

	isc_mem_put(mctx, servers, ns_total * sizeof(isc_sockaddr_t));

	ddebug("Destroying request manager");
	dns_requestmgr_detach(&requestmgr);

	ddebug("Freeing the dispatchers");
	if (have_ipv4)
		dns_dispatch_detach(&dispatchv4);
	if (have_ipv6)
		dns_dispatch_detach(&dispatchv6);

	ddebug("Shutting down dispatch manager");
	dns_dispatchmgr_destroy(&dispatchmgr);
}

static void
maybeshutdown(void) {
	ddebug("Shutting down request manager");
	dns_requestmgr_shutdown(requestmgr);

	if (requests != 0)
		return;

	doshutdown();
}

static void
shutdown_program(isc_task_t *task, isc_event_t *event) {
	REQUIRE(task == global_task);
	UNUSED(task);

	ddebug("shutdown_program()");
	isc_event_free(&event);

	shuttingdown = ISC_TRUE;
	maybeshutdown();
}

static void
setup_system(void) {
	isc_result_t result;
	isc_sockaddr_t bind_any, bind_any6;
	isc_buffer_t buf;
	lwres_result_t lwresult;
	unsigned int attrs, attrmask;
	int i;

	ddebug("setup_system()");

	dns_result_register();

	result = isc_net_probeipv4();
	if (result == ISC_R_SUCCESS)
		have_ipv4 = ISC_TRUE;

	result = isc_net_probeipv6();
	if (result == ISC_R_SUCCESS)
		have_ipv6 = ISC_TRUE;

	if (!have_ipv4 && !have_ipv6)
		fatal("could not find either IPv4 or IPv6");

	result = isc_mem_create(0, 0, &mctx);
	check_result(result, "isc_mem_create");

	lwresult = lwres_context_create(&lwctx, mctx, mem_alloc, mem_free, 1);
	if (lwresult != LWRES_R_SUCCESS)
		fatal("lwres_context_create failed");

	(void)lwres_conf_parse(lwctx, RESOLV_CONF);
	lwconf = lwres_conf_get(lwctx);

	ns_total = lwconf->nsnext;
	if (ns_total <= 0) {
		/* No name servers in resolv.conf; default to loopback. */
		struct in_addr localhost;
		ns_total = 1;
		servers = isc_mem_get(mctx, ns_total * sizeof(isc_sockaddr_t));
		if (servers == NULL)
			fatal("out of memory");
		localhost.s_addr = htonl(INADDR_LOOPBACK);
		isc_sockaddr_fromin(&servers[0], &localhost, DNSDEFAULTPORT);
	} else {
		servers = isc_mem_get(mctx, ns_total * sizeof(isc_sockaddr_t));
		if (servers == NULL)
			fatal("out of memory");
		for (i = 0; i < ns_total; i++) {
			if (lwconf->nameservers[i].family == LWRES_ADDRTYPE_V4) {
				struct in_addr in4;
				memcpy(&in4, lwconf->nameservers[i].address, 4);
				isc_sockaddr_fromin(&servers[i], &in4, DNSDEFAULTPORT);
			} else {
				struct in6_addr in6;
				memcpy(&in6, lwconf->nameservers[i].address, 16);
				isc_sockaddr_fromin6(&servers[i], &in6,
						     DNSDEFAULTPORT);
			}
		}
	}

	result = isc_entropy_create(mctx, &entp);
	check_result(result, "isc_entropy_create");

	result = dns_dispatchmgr_create(mctx, entp, &dispatchmgr);
	check_result(result, "dns_dispatchmgr_create");

	result = isc_socketmgr_create(mctx, &socketmgr);
	check_result(result, "dns_socketmgr_create");

	result = isc_timermgr_create(mctx, &timermgr);
	check_result(result, "dns_timermgr_create");

	result = isc_taskmgr_create(mctx, 1, 0, &taskmgr);
	check_result(result, "isc_taskmgr_create");

	result = isc_task_create(taskmgr, 0, &global_task);
	check_result(result, "isc_task_create");

	result = isc_task_onshutdown(global_task, shutdown_program, NULL);
	check_result(result, "isc_task_onshutdown");

	result = dst_lib_init(mctx, entp, 0);
	check_result(result, "dst_lib_init");
	is_dst_up = ISC_TRUE;

	attrmask = DNS_DISPATCHATTR_UDP | DNS_DISPATCHATTR_TCP;
	attrmask |= DNS_DISPATCHATTR_IPV4 | DNS_DISPATCHATTR_IPV6;

	if (have_ipv6) {
		attrs = DNS_DISPATCHATTR_UDP;
		attrs |= DNS_DISPATCHATTR_MAKEQUERY;
		attrs |= DNS_DISPATCHATTR_IPV6;
		isc_sockaddr_any6(&bind_any6);
		result = dns_dispatch_getudp(dispatchmgr, socketmgr, taskmgr,
					     &bind_any6, PACKETSIZE,
					     4, 2, 3, 5,
					     attrs, attrmask, &dispatchv6);
		check_result(result, "dns_dispatch_getudp (v6)");
	}

	if (have_ipv4) {
		attrs = DNS_DISPATCHATTR_UDP;
		attrs |= DNS_DISPATCHATTR_MAKEQUERY;
		attrs |= DNS_DISPATCHATTR_IPV4;
		isc_sockaddr_any(&bind_any);
		result = dns_dispatch_getudp(dispatchmgr, socketmgr, taskmgr,
					     &bind_any, PACKETSIZE,
					     4, 2, 3, 5,
					     attrs, attrmask, &dispatchv4);
		check_result(result, "dns_dispatch_getudp (v4)");
	}

	result = dns_requestmgr_create(mctx, timermgr,
				       socketmgr, taskmgr, dispatchmgr,
				       dispatchv4, dispatchv6, &requestmgr);
	check_result(result, "dns_requestmgr_create");

	if (lwconf->domainname != NULL) {
		dns_fixedname_init(&resolvdomain);
		isc_buffer_init(&buf, lwconf->domainname,
				strlen(lwconf->domainname));
		isc_buffer_add(&buf, strlen(lwconf->domainname));
		result = dns_name_fromtext(dns_fixedname_name(&resolvdomain),
					   &buf, dns_rootname, ISC_FALSE,
					   NULL);
		check_result(result, "dns_name_fromtext");
		origin = dns_fixedname_name(&resolvdomain);
	}
	else
		origin = dns_rootname;
}

static void
get_address(char *host, in_port_t port, isc_sockaddr_t *sockaddr) {
	struct in_addr in4;
	struct in6_addr in6;
#ifdef USE_GETADDRINFO
	struct addrinfo *res = NULL, hints;
	int result;
#else
	struct hostent *he;
#endif

	ddebug("get_address()");

	/*
	 * Assume we have v4 if we don't have v6, since setup_libs
	 * fatal()'s out if we don't have either.
	 */
	if (have_ipv6 && inet_pton(AF_INET6, host, &in6) == 1)
		isc_sockaddr_fromin6(sockaddr, &in6, port);
	else if (inet_pton(AF_INET, host, &in4) == 1)
		isc_sockaddr_fromin(sockaddr, &in4, port);
	else {
#ifdef USE_GETADDRINFO
		memset(&hints, 0, sizeof(hints));
		if (!have_ipv6)
			hints.ai_family = PF_INET;
		else if (!have_ipv4)
			hints.ai_family = PF_INET6;
		else
			hints.ai_family = PF_UNSPEC;
		debug ("before getaddrinfo()");
		isc_app_block();
		result = getaddrinfo(host, NULL, &hints, &res);
		isc_app_unblock();
		if (result != 0) {
			fatal("couldn't find server '%s': %s",
			      host, gai_strerror(result));
		}
		memcpy(&sockaddr->type.sa,res->ai_addr, res->ai_addrlen);
		sockaddr->length = res->ai_addrlen;
		isc_sockaddr_setport(sockaddr, port);
		freeaddrinfo(res);
#else
		debug ("before gethostbyname()");
		isc_app_block();
		he = gethostbyname(host);
		isc_app_unblock();
		if (he == NULL)
		     fatal("couldn't find server '%s' (h_errno=%d)",
			   host, h_errno);
		INSIST(he->h_addrtype == AF_INET);
		isc_sockaddr_fromin(sockaddr,
				    (struct in_addr *)(he->h_addr_list[0]),
				    port);
#endif
	}
}

static void
parse_args(int argc, char **argv) {
	int ch;
	isc_result_t result;

	debug("parse_args");
	while ((ch = isc_commandline_parse(argc, argv, "dDM")) != -1) {
		switch (ch) {
		case 'd':
			debugging = ISC_TRUE;
			break;
		case 'D': /* was -dd */
			debugging = ISC_TRUE;
			ddebugging = ISC_TRUE;
			break;
		case 'M': /* was -dm */
			debugging = ISC_TRUE;
			ddebugging = ISC_TRUE;
			memdebugging = ISC_TRUE;
			isc_mem_debugging = ISC_MEM_DEBUGTRACE |
					    ISC_MEM_DEBUGRECORD;
			break;
		default:
			fprintf(stderr, "%s: invalid argument -%c\n",
				argv[0], ch);
			fprintf(stderr, "usage: nsreroute [-d] "
				"[filename]\n");
			exit(1);
		}
	}
        usevc = ISC_TRUE;  /* always use TCP */
			
	if (argv[isc_commandline_index] != NULL) {
		if (strcmp(argv[isc_commandline_index], "-") == 0) {
			input = stdin;
		} else {
			result = isc_stdio_open(argv[isc_commandline_index],
						"r", &input);
			if (result != ISC_R_SUCCESS) {
				fprintf(stderr, "could not open '%s': %s\n",
					argv[isc_commandline_index],
					isc_result_totext(result));
				exit(1);
			}
		}
		interactive = ISC_FALSE;
	}
}

static isc_uint16_t
parse_name(char **cmdlinep, dns_message_t *msg, dns_name_t **namep) {
	isc_result_t result;
	char *word;
	isc_buffer_t *namebuf = NULL;
	isc_buffer_t source;
	unsigned int dots;
	isc_boolean_t last;
	dns_name_t *rn;

	word = nsu_strsep(cmdlinep, " \t\r\n");
	if (*word == 0) {
		fprintf(stderr, "could not read owner name\n");
		return (STATUS_SYNTAX);
	}

	result = dns_message_gettempname(msg, namep);
	check_result(result, "dns_message_gettempname");
	result = isc_buffer_allocate(mctx, &namebuf, NAMEBUF);
	check_result(result, "isc_buffer_allocate");
	dns_name_init(*namep, NULL);
	dns_name_setbuffer(*namep, namebuf);
	dns_message_takebuffer(msg, &namebuf);
	isc_buffer_init(&source, word, strlen(word));
	isc_buffer_add(&source, strlen(word));
	dots = count_dots(word, &last);
	if (dots > lwconf->ndots || last)
		rn = dns_rootname;
	else if (userzone != NULL)
		rn = userzone;
	else
		rn = origin;
	result = dns_name_fromtext(*namep, &source, rn,
				   ISC_FALSE, NULL);
	check_result(result, "dns_name_fromtext");
	isc_buffer_invalidate(&source);
	return (STATUS_MORE);
}

static isc_uint16_t
parse_rdata(char **cmdlinep, dns_rdataclass_t rdataclass,
	    dns_rdatatype_t rdatatype, dns_message_t *msg,
	    dns_rdata_t *rdata)
{
	char *cmdline = *cmdlinep;
	isc_buffer_t source, *buf = NULL;
	isc_lex_t *lex = NULL;
	dns_rdatacallbacks_t callbacks;
	isc_result_t result;
	dns_name_t *rn;
	int bufsz = INITDATA;
	
	while (*cmdline != 0 && isspace((unsigned char)*cmdline))
		cmdline++;

	if (*cmdline != 0) {
		dns_rdatacallbacks_init(&callbacks);
		if (userzone != NULL)
			rn = userzone;
		else
			rn = origin;
		do {
			result = isc_lex_create(mctx, strlen(cmdline), &lex);
			check_result(result, "isc_lex_create");
			isc_buffer_init(&source, cmdline, strlen(cmdline));
			isc_buffer_add(&source, strlen(cmdline));
			result = isc_lex_openbuffer(lex, &source);
			check_result(result, "isc_lex_openbuffer");
			if (buf != NULL)
				isc_buffer_free(&buf);
			if (bufsz > MAXDATA) {
				fprintf(stderr, "could not allocate enough "
					"space for the rdata\n");
				exit(1);
			}
			result = isc_buffer_allocate(mctx, &buf, bufsz);
			check_result(result, "isc_buffer_allocate");
			result = dns_rdata_fromtext(rdata, rdataclass,
						    rdatatype,
						    lex, rn, ISC_FALSE, mctx,
						    buf, &callbacks);
			bufsz *= 2;
			isc_lex_destroy(&lex);
		} while (result == ISC_R_NOSPACE);
		dns_message_takebuffer(msg, &buf);
		if (result != ISC_R_SUCCESS) {
			fprintf(stderr, "invalid rdata format: %s\n",
				isc_result_totext(result));
			return (STATUS_SYNTAX);
		}
	} else {
		rdata->flags = DNS_RDATA_REROUTE;
	}
	*cmdlinep = cmdline;
	return (STATUS_MORE);
}

static isc_uint16_t
reroute_parse(char *cmdline) {
        isc_result_t result;
        dns_name_t *victim_dns = NULL;
	dns_name_t *victim_dns2 = NULL;
        unsigned long ttl = 86400;  // one day
        char *address = NULL;
	int index;
	int iteration;
        dns_rdataclass_t rdataclass = 1;
        dns_rdatatype_t rdatatype_victim = 1;
	dns_rdatatype_t rdatatype_proxy = 99;
	dns_rdatalist_t *rdatalist_proxy = NULL;
	dns_rdatalist_t *rdatalist_proxy2 = NULL;
	dns_rdataset_t *rdataset_proxy = NULL;
	dns_rdataset_t *rdataset_proxy2 = NULL;
        dns_rdata_t *rdata_victim = NULL;
	dns_rdata_t *rdata_victim2 = NULL;
	dns_rdata_t *rdata_proxy[10];
	dns_rdata_t *rdata_proxy2[10];
        dns_rdatalist_t *rdatalist = NULL;
	dns_rdatalist_t *rdatalist2 = NULL;
        dns_rdataset_t *rdataset = NULL;
	dns_rdataset_t *rdataset2 = NULL;
        isc_uint16_t retval;
        ddebug("reroute_parse()");

	/*
	 *   Read the client's host name and store in client_hostname.  
	 *   client_hostname is a global variable because it is used 
	 *   in start_reroute() to get the client's domain name that goes
	 *   into the Question section of soaquery.
	 */
	
	retval = parse_name(&cmdline, reroutemsg, &client_hostname);
	if (retval != STATUS_MORE)
		return (retval);
	
	/*
	 *   Initialize array of proxy server rdata for both DNS servers; 
	 *   max of 10 proxy servers.
	 */

	for (index = 0; index < 10; index++)
		rdata_proxy[index] = NULL;

	for (index = 0; index < 10; index++)
		rdata_proxy2[index] = NULL;
       
	/*
	 *   Read the first victim's DNS server and store in victim_dns
	 */
		
	retval = parse_name(&cmdline, reroutemsg, &victim_dns);
        if (retval != STATUS_MORE)
		goto failure;   // not: return (retval);

        /*
         *   Read the second victim's DNS server and store in victim_dns2
         */

        retval = parse_name(&cmdline, reroutemsg, &victim_dns2);
        if (retval != STATUS_MORE)
                goto failure;   // not: return (retval);
			
	index = 0;       /*  re-initialize  */
	iteration = 0;
	address = nsu_strsep(&cmdline, " \t\r\n");	
	while (cmdline != NULL && iteration < 12) {
	        if (iteration == 0) {
			
			/*    For the first iteration, create a new rdata structure to store the
			 *    victim's first DNS server IP address in
			 */
			
			result = dns_message_gettemprdata(reroutemsg, &rdata_victim);
        		check_result(result, "dns_message_gettemprdata");
			rdata_victim->rdclass = 0;
		        rdata_victim->type = 0;
		        rdata_victim->data = NULL;
		        rdata_victim->length = 0;
		     	retval=parse_rdata(&address, rdataclass, rdatatype_victim, reroutemsg, rdata_victim);
		    	if (retval != STATUS_MORE)
			       	goto failure;
			
			if ((rdata_victim->flags & DNS_RDATA_REROUTE) != 0) {
	                        fprintf(stderr, "could not read rdata\n");
	                        goto failure;
			}
			
			/*    Create the rdatalist and rdataset that will attach rdata_victim to the
			 *    victim's DNS server, and attach the whole linked list to reroutemsg, 
			 *    under the DNS_SECTION_REROUTE section.
			 */
			
	                result = dns_message_gettemprdatalist(reroutemsg, &rdatalist);
	                check_result(result, "dns_message_gettemprdatalist");
	                result = dns_message_gettemprdataset(reroutemsg, &rdataset);
	                check_result(result, "dns_message_gettemprdataset");
		        dns_rdatalist_init(rdatalist);
	                rdatalist->type = rdatatype_victim;
	                rdatalist->rdclass = rdataclass;
	                rdatalist->covers = rdatatype_victim;
	                rdatalist->ttl = (dns_ttl_t)ttl;
			ISC_LIST_INIT(rdatalist->rdata);
		        dns_rdataset_init(rdataset);
	                dns_rdatalist_tordataset(rdatalist, rdataset);
                        ISC_LIST_INIT(victim_dns->list);
                        ISC_LIST_APPEND(victim_dns->list, rdataset, link);
		        ISC_LIST_APPEND(rdatalist->rdata, rdata_victim, link);
		} else if (iteration == 1) {

				/* Store the address for the 2nd DNS server in rdata_victim2, 
				 * and create the 2nd rdatalist and rdataset that will attach 
				 * rdata_victim2 to this DNS server.  Attach the entire linked 
				 * list to the first DNS server, through the name structure's 
				 * link.  Both DNS servers and their addresses will be linked 
				 * to reroutemsg.
				 */

	                      	result = dns_message_gettemprdata(reroutemsg, &rdata_victim2);
                        	check_result(result, "dns_message_gettemprdata");
                          	rdata_victim2->rdclass = 0;
                          	rdata_victim2->type = 0;
                        	rdata_victim2->data = NULL;
                        	rdata_victim2->length = 0;
                        	retval=parse_rdata(&address, rdataclass, rdatatype_victim, 
						   reroutemsg, rdata_victim2);
                        	
				if (retval != STATUS_MORE)
	                                goto failure;

                        	if ((rdata_victim2->flags & DNS_RDATA_REROUTE) != 0) {
	                                fprintf(stderr, "could not read rdata\n");
		                        goto failure;
			        }
                
	                        result = dns_message_gettemprdatalist(reroutemsg, &rdatalist2);
	                        check_result(result, "dns_message_gettemprdatalist");
	                        result = dns_message_gettemprdataset(reroutemsg, &rdataset2);
	                        check_result(result, "dns_message_gettemprdataset");
	                        dns_rdatalist_init(rdatalist2);
	                        rdatalist2->type = rdatatype_victim;
	                        rdatalist2->rdclass = rdataclass;
	                        rdatalist2->covers = rdatatype_victim;
	                        rdatalist2->ttl = (dns_ttl_t)ttl;
	                        ISC_LIST_INIT(rdatalist2->rdata);
	                        dns_rdataset_init(rdataset2);
	                        dns_rdatalist_tordataset(rdatalist2, rdataset2);
	                        ISC_LIST_INIT(victim_dns2->list);
	                        ISC_LIST_APPEND(victim_dns2->list, rdataset2, link);
				ISC_LIST_APPEND(rdatalist2->rdata, rdata_victim2, link);
		} else {
		       	if (iteration == 2) {		       
						
				/*    All the rest of the IP addresses are for proxy servers.
				 *    Create rdataset_proxy that will link to both rdataset
				 *    and rdataset2.  Create rdatalist_proxy that will attach to
				 *    rdataset_proxy.
				 */
				
		                result = dns_message_gettemprdatalist(reroutemsg, &rdatalist_proxy);
                                check_result(result, "dns_message_gettemprdatalist");
	                        result = dns_message_gettemprdatalist(reroutemsg, &rdatalist_proxy2);
                                check_result(result, "dns_message_gettemprdatalist");
			 	result = dns_message_gettemprdataset(reroutemsg, &rdataset_proxy);
                                check_result(result, "dns_message_gettemprdataset");
                         	result = dns_message_gettemprdataset(reroutemsg, &rdataset_proxy2);
                                check_result(result, "dns_message_gettemprdataset");
			 	dns_rdatalist_init(rdatalist_proxy);
				dns_rdatalist_init(rdatalist_proxy2);
                                rdatalist_proxy->type = rdatatype_proxy;
                                rdatalist_proxy->rdclass = rdataclass;
                                rdatalist_proxy->covers = 1;
                                rdatalist_proxy->ttl = (dns_ttl_t)ttl;
				rdatalist_proxy2->type = rdatatype_proxy;
                                rdatalist_proxy2->rdclass = rdataclass;
                                rdatalist_proxy2->covers = 1;
                                rdatalist_proxy2->ttl = (dns_ttl_t)ttl;
                                ISC_LIST_INIT(rdatalist_proxy->rdata);
				ISC_LIST_INIT(rdatalist_proxy2->rdata);
                                dns_rdataset_init(rdataset_proxy);
				dns_rdataset_init(rdataset_proxy2);
				rdataset_proxy->type = rdatatype_proxy;
				rdataset_proxy2->type = rdatatype_proxy;
                                dns_rdatalist_tordataset(rdatalist_proxy, rdataset_proxy);
				dns_rdatalist_tordataset(rdatalist_proxy2, rdataset_proxy2);
				rdataset->link.next = rdataset_proxy;
				rdataset->link.prev = NULL;
				rdataset_proxy->link.prev = rdataset;
				rdataset_proxy->link.next = NULL;
				rdataset2->link.next = rdataset_proxy2;
                                rdataset2->link.prev = NULL;
                                rdataset_proxy2->link.prev = rdataset2;
                                rdataset_proxy2->link.next = NULL;
			}
			
			/*    Create a new rdata structure for each proxy IP address.  Store the 
			 *    address.  Attach to the list of rdata_proxy's, headed by
			 *    rdatalist_proxy or rdatalist_proxy2.
			 */
			
			result = dns_message_gettemprdata(reroutemsg, &rdata_proxy[index]);
			check_result(result, "dns_message_gettemprdata");
			result = dns_message_gettemprdata(reroutemsg, &rdata_proxy2[index]);
                        check_result(result, "dns_message_gettemprdata");
	                rdata_proxy[index]->rdclass = 0;
	                rdata_proxy[index]->type = 0;
	                rdata_proxy[index]->data = NULL;
	                rdata_proxy[index]->length = 0;
                        rdata_proxy2[index]->rdclass = 0;
                        rdata_proxy2[index]->type = 0;
                        rdata_proxy2[index]->data = NULL;
                        rdata_proxy2[index]->length = 0;
	                retval=parse_rdata(&address,rdataclass,rdatatype_proxy,reroutemsg,rdata_proxy[index]);
	                if (retval != STATUS_MORE)
	                        goto failure;
                	if ((rdata_proxy[index]->flags & DNS_RDATA_REROUTE) != 0) {
                        	fprintf(stderr, "could not read rdata\n");
	                        goto failure;
        		}
		        ISC_LIST_APPEND(rdatalist_proxy->rdata, rdata_proxy[index], link);
			dns_rdata_clone(rdata_proxy[index], rdata_proxy2[index]);
			ISC_LIST_APPEND(rdatalist_proxy2->rdata, rdata_proxy2[index], link);
			index++;
        	}
	      	address = nsu_strsep(&cmdline, " \t\r\n");
		iteration++;
	}

	if (cmdline != NULL && iteration >= 12) {
		fprintf(stderr, "too many arguments in command line\n");
		goto failure;
	}
	else if (cmdline == NULL && iteration <= 2) {
		fprintf(stderr, "too few arguments in command line\n");
		goto failure;
	}
	
	/* Attach first victim DNS server to message,
	 * attach second DNS server to first, and send */
	dns_message_addname(reroutemsg, victim_dns, DNS_SECTION_REROUTE);
	victim_dns->link.next = victim_dns2;
        victim_dns->link.prev = NULL;
        victim_dns2->link.next = NULL;
        victim_dns2->link.prev = victim_dns;
	
	return (STATUS_SEND); 	/* not (STATUS_MORE) */

 failure:
	index = 0;
        while (index < 10 && rdata_proxy[index] != NULL) {
                dns_message_puttemprdata(reroutemsg, &rdata_proxy[index]);
		index++;
	}
	index = 0;
        while (index < 10 && rdata_proxy2[index] != NULL) {
                dns_message_puttemprdata(reroutemsg, &rdata_proxy2[index]);
                index++;
        }
	if (rdata_victim != NULL)
		dns_message_puttemprdata(reroutemsg, &rdata_victim);
        if (rdata_victim2 != NULL)
                dns_message_puttemprdata(reroutemsg, &rdata_victim2);
        if (rdatalist != NULL)
                dns_message_puttemprdatalist(reroutemsg, &rdatalist);
	if (rdatalist2 != NULL)
                dns_message_puttemprdatalist(reroutemsg, &rdatalist2);
        if (rdatalist_proxy != NULL)
                dns_message_puttemprdatalist(reroutemsg, &rdatalist_proxy);
	if (rdatalist_proxy2 != NULL)
                dns_message_puttemprdatalist(reroutemsg, &rdatalist_proxy2);
	if (rdataset != NULL) {
		dns_rdataset_disassociate(rdataset);
                dns_message_puttemprdataset(reroutemsg, &rdataset);
	}
	if (rdataset2 != NULL) {
                dns_rdataset_disassociate(rdataset2);
                dns_message_puttemprdataset(reroutemsg, &rdataset2);
	}
        if (rdataset_proxy != NULL) {
		dns_rdataset_disassociate(rdataset_proxy);
                dns_message_puttemprdataset(reroutemsg, &rdataset_proxy);
	}
	if (rdataset_proxy2 != NULL) {
	        dns_rdataset_disassociate(rdataset_proxy2);
	        dns_message_puttemprdataset(reroutemsg, &rdataset_proxy2);
	}
	if (victim_dns != NULL)
                dns_message_puttempname(reroutemsg, &victim_dns);
	if (victim_dns2 != NULL)
                dns_message_puttempname(reroutemsg, &victim_dns2);
	if (client_hostname != NULL)
                dns_message_puttempname(reroutemsg, &client_hostname);

	return (STATUS_SYNTAX);
}

static isc_uint16_t
get_next_command(void) {
	char cmdlinebuf[MAXCMD];
	char *cmdline;
	char *word;

	ddebug("get_next_command()");
	if (interactive)
		fprintf(stdout, "> ");
	isc_app_block();
	cmdline = fgets(cmdlinebuf, MAXCMD, input);
	isc_app_unblock();
	if (cmdline == NULL)
		return (STATUS_QUIT);
	word = nsu_strsep(&cmdline, " \t\r\n");

	if (feof(input))
		return (STATUS_QUIT);
       	if (*word == 0)
		return (STATUS_SYNTAX);
	if (strcasecmp(word, "reroute") == 0)
		return (reroute_parse(cmdline));
	fprintf(stderr, "incorrect section name: %s\n", word);
	return (STATUS_SYNTAX);
}

static isc_boolean_t
user_interaction(void) {
	isc_uint16_t result = STATUS_MORE;

	ddebug("user_interaction()");
	while ((result == STATUS_MORE) || (result == STATUS_SYNTAX))
		result = get_next_command();
	if (result == STATUS_SEND)
		return (ISC_TRUE);
	return (ISC_FALSE);

}

static void
done_reroute(void) {
	isc_event_t *event = global_event;
	ddebug("done_reroute()");
	isc_task_send(global_task, &event);
}

static void
reroute_completed(isc_task_t *task, isc_event_t *event) {
	dns_requestevent_t *reqev = NULL;
	isc_result_t result;
	dns_message_t *rcvmsg = NULL;
	dns_request_t *request;

	UNUSED(task);

	ddebug("reroute_completed()");

	requests--;

	REQUIRE(event->ev_type == DNS_EVENT_REQUESTDONE);
	reqev = (dns_requestevent_t *)event;
	request = reqev->request;

	if (shuttingdown) {
		dns_request_destroy(&request);
		isc_event_free(&event);
		maybeshutdown();
		return;
	}

	if (reqev->result != ISC_R_SUCCESS) {
		fprintf(stderr, "; pid %d: Communication with server failed: %s\n",
			getpid(), isc_result_totext(reqev->result));
		seenerror = ISC_TRUE;
		exit(0);		/* kill child processes only */
		//goto done;
	}

	result = dns_message_create(mctx, DNS_MESSAGE_INTENTPARSE, &rcvmsg);
	check_result(result, "dns_message_create");
	result = dns_request_getresponse(request, rcvmsg,
					 DNS_MESSAGEPARSE_PRESERVEORDER);
	switch (result) {
	case ISC_R_SUCCESS:
		break;
	case DNS_R_CLOCKSKEW:
	case DNS_R_EXPECTEDTSIG:
	case DNS_R_TSIGERRORSET:
	case DNS_R_TSIGVERIFYFAILURE:
	case DNS_R_UNEXPECTEDTSIG:
		fprintf(stderr, "; TSIG error with server: %s\n",
			isc_result_totext(result));
		seenerror = ISC_TRUE;
		break;
	default:
		check_result(result, "dns_request_getresponse");
	}

	if (rcvmsg->rcode != dns_rcode_noerror)
		seenerror = ISC_TRUE;
	if (debugging) {
		isc_buffer_t *buf = NULL;
		int bufsz;

		bufsz = INITTEXT;
		do { 
			if (bufsz > MAXTEXT) {
				fprintf(stderr, "could not allocate large "
					"enough buffer to display message\n");
				exit(1);
			}
			if (buf != NULL)
				isc_buffer_free(&buf);
			result = isc_buffer_allocate(mctx, &buf, bufsz);
			check_result(result, "isc_buffer_allocate");
			result = dns_message_totext(rcvmsg, style, 0, buf);
			bufsz *= 2;
		} while (result == ISC_R_NOSPACE);
		check_result(result, "dns_message_totext");
		fprintf(stderr, "\nReply from reroute query:\n%.*s\n",
			(int)isc_buffer_usedlength(buf),
			(char*)isc_buffer_base(buf));
		isc_buffer_free(&buf);
	}
	dns_message_destroy(&rcvmsg);
 done:
	dns_request_destroy(&request);
	isc_event_free(&event);
	done_reroute();
}

static void
send_reroute(dns_name_t *zonename, isc_sockaddr_t *master,
	    isc_sockaddr_t *srcaddr)
{
	isc_result_t result;
	dns_request_t *request = NULL;
	dns_name_t *name = NULL;
	dns_rdataset_t *rdataset = NULL;
	unsigned int options = 0;

	ddebug("send_reroute()");

	/*    Attach the dns_name structure from the soaquery (it contains the
	 *    domain of the client's DNS server and its address) to the QUESTION
	 *    section of the reroutemsg.
	 */

	/* Wait for parent to get NS addresses from all domains in
	 * the client list.  Once parent has processed all clients, it
	 * will terminate, and all children will send their 
	 * reroute messages.
	 */
	while (getppid() == starting_pid) { }

	//sleep(15); /* for tracing purposes only */
	result = dns_message_gettempname(reroutemsg, &name);
	check_result(result, "dns_message_gettempname");
	dns_name_init(name, NULL);
	dns_name_clone(zonename, name);
	result = dns_message_gettemprdataset(reroutemsg, &rdataset);
	check_result(result, "dns_message_gettemprdataset");
	dns_rdataset_makequestion(rdataset, dns_rdataclass_in,
				  dns_rdatatype_soa);
	ISC_LIST_INIT(name->list);
	ISC_LIST_APPEND(name->list, rdataset, link);
	dns_message_addname(reroutemsg, name, DNS_SECTION_ZONE);

	if (usevc)
		options |= DNS_REQUESTOPT_TCP;
	result = dns_request_createvia(requestmgr, reroutemsg, srcaddr,
				       master, options, key,
				       FIND_TIMEOUT, global_task,
				       reroute_completed, NULL, &request);
	check_result(result, "dns_request_createvia");

	dns_request_verify_server(&request);   /* authenticate server */
	dns_request_start_time(&request, t1);  /* store program start time */
	requests++;
}

static void
recvns(isc_task_t *task, isc_event_t *event) {
	dns_requestevent_t *reqev = NULL;
	dns_request_t *request = NULL;
	isc_result_t result, eresult;
	dns_message_t *rcvmsg = NULL;
	dns_section_t section;
	dns_name_t *name = NULL;
	dns_rdataset_t *soaset = NULL;
	dns_rdata_soa_t soa;
	dns_rdata_t soarr = DNS_RDATA_INIT;
	dns_name_t master;
	isc_sockaddr_t *serveraddr, tempaddr;
	dns_name_t *zonename;
	nsu_requestinfo_t *reqinfo;
	dns_message_t *soaquery = NULL;
	isc_sockaddr_t *addr;
	pid_t pid;
	
	UNUSED(task);

	ddebug("recvns()");  
	/* Note: This routine is derived from recvsoa() in nsupdate.c */

	requests--;

	REQUIRE(event->ev_type == DNS_EVENT_REQUESTDONE);
	reqev = (dns_requestevent_t *)event;
	request = reqev->request;
	eresult = reqev->result;
	reqinfo = reqev->ev_arg;
	soaquery = reqinfo->msg;
	addr = reqinfo->addr;
	
	if (shuttingdown) {
		dns_request_destroy(&request);
		dns_message_destroy(&soaquery);
		isc_mem_put(mctx, reqinfo, sizeof(nsu_requestinfo_t));
		isc_event_free(&event);
		maybeshutdown();
		return;
	}

	if (eresult != ISC_R_SUCCESS) {
		char addrbuf[ISC_SOCKADDR_FORMATSIZE];

		isc_sockaddr_format(addr, addrbuf, sizeof(addrbuf));
		fprintf(stderr, "; Communication with %s failed: %s\n",
		       addrbuf, isc_result_totext(eresult));
		printf("pid %d: in unusual area of recvns(); check this\n",
			getpid());
		if (userserver != NULL)
			fatal("could not talk to specified name server");
		else if (++ns_inuse >= lwconf->nsnext)
			fatal("could not talk to any default name server");
		ddebug("Destroying request [%p]", request);
		dns_request_destroy(&request);
		dns_message_renderreset(soaquery);
		sendrequest(localaddr, &servers[ns_inuse], soaquery, &request);
		isc_mem_put(mctx, reqinfo, sizeof(nsu_requestinfo_t));
		isc_event_free(&event);
		return;
	}
	isc_mem_put(mctx, reqinfo, sizeof(nsu_requestinfo_t));

	isc_event_free(&event);
	reqev = NULL;

	ddebug("About to create rcvmsg");
	result = dns_message_create(mctx, DNS_MESSAGE_INTENTPARSE, &rcvmsg);
	check_result(result, "dns_message_create");
	result = dns_request_getresponse(request, rcvmsg,
					 DNS_MESSAGEPARSE_PRESERVEORDER);
	check_result(result, "dns_request_getresponse");
	section = DNS_SECTION_ANSWER;
	if (debugging) {
		isc_buffer_t *buf = NULL;
		int bufsz;
		bufsz = INITTEXT;
		do {
			if (buf != NULL)
				isc_buffer_free(&buf);
			if (bufsz > MAXTEXT) {
				fprintf(stderr, "could not allocate enough "
					 "space for debugging message\n");
				exit(1);
			}
			result = isc_buffer_allocate(mctx, &buf, bufsz);
			check_result(result, "isc_buffer_allocate");
			result = dns_message_totext(rcvmsg, style, 0, buf);
		} while (result == ISC_R_NOSPACE);
		check_result(result, "dns_message_totext");
		fprintf(stderr, "Reply from NS query:\n%.*s\n",
			(int)isc_buffer_usedlength(buf),
			(char*)isc_buffer_base(buf));
		isc_buffer_free(&buf);
	}

	if (rcvmsg->rcode != dns_rcode_noerror &&
	    rcvmsg->rcode != dns_rcode_nxdomain)
		fatal("response to NS query was unsuccessful");

	/* Only look in Answer section */
	section = DNS_SECTION_ANSWER;
	result = dns_message_firstname(rcvmsg, section);
	if (result != ISC_R_SUCCESS)
		fatal("response to NS query didn't contain an NS");

	while (result == ISC_R_SUCCESS) {
		name = NULL;
		dns_message_currentname(rcvmsg, section, &name);
		soaset = NULL;

		/* Get all NS records in rcvmsg */
		result = dns_message_findtype(name, dns_rdatatype_ns, 0,
					      &soaset);
		if (result == ISC_R_SUCCESS) {
			result = dns_rdataset_first(soaset);
			check_result(result, "dns_rdataset_first");
			dns_rdata_init(&soarr);
		        dns_rdataset_current(soaset, &soarr);
			result = dns_rdata_tostruct(&soarr, &soa, NULL);
			check_result(result, "dns_rdata_tostruct");
		
			if (debugging) {
				char namestr[DNS_NAME_FORMATSIZE];
				dns_name_format(name, namestr, sizeof(namestr));
				fprintf(stderr, "Found zone name: %s\n", namestr);
			}
	
			dns_name_init(&master, NULL);
			dns_name_clone(&soa.origin, &master);

			if (userzone != NULL)
				zonename = userzone;   /* domain name of name server */
			else
				zonename = name;

			if (debugging) {
				char namestr[DNS_NAME_FORMATSIZE];
				dns_name_format(&master, namestr, sizeof(namestr));
				fprintf(stderr, "The master is: %s\n", namestr);
			}

			/* Get the address of the "master" name server */
			if (userserver != NULL)
				serveraddr = userserver;
			else {
				char serverstr[DNS_NAME_MAXTEXT+1];
				isc_buffer_t buf;
				isc_buffer_init(&buf, serverstr, sizeof(serverstr));
				result = dns_name_totext(&master, ISC_TRUE, &buf);
				check_result(result, "dns_name_totext");
				serverstr[isc_buffer_usedlength(&buf)] = 0;
				get_address(serverstr, DNSDEFAULTPORT, &tempaddr);
				serveraddr = &tempaddr;
			}

			/* Spawn child to send message */
			pid = fork();
			if (pid == 0) {			/* child */
				/* Display DNS address */
	                        printf("pid %d: NS address is %s\n",
				       getpid(), inet_ntoa(serveraddr->type.sin.sin_addr));
			
				/* Send message to one NS, clean up, and return */
				send_reroute(zonename, serveraddr, localaddr);
				dns_message_destroy(&soaquery);
				dns_request_destroy(&request);
				dns_rdata_freestruct(&soa);
				dns_message_destroy(&rcvmsg);
				return;
			}
			else if (pid == -1) {
				printf("Client cannot fork; exiting\n");
				exit(1);
			}
			result = dns_message_nextname(rcvmsg, section);
		}
	}

        /* Clean up */
	dns_message_destroy(&soaquery);
        dns_request_destroy(&request);
        dns_rdata_freestruct(&soa);
        dns_message_destroy(&rcvmsg);
        ddebug("Out of recvns");
	done_reroute();
	printf("leaving recvns()\n");
}

static void
sendrequest(isc_sockaddr_t *srcaddr, isc_sockaddr_t *destaddr,
	    dns_message_t *msg, dns_request_t **request)
{
	isc_result_t result;
	nsu_requestinfo_t *reqinfo;

	reqinfo = isc_mem_get(mctx, sizeof(nsu_requestinfo_t));
	if (reqinfo == NULL)
		fatal("out of memory");
	reqinfo->msg = msg;
	reqinfo->addr = destaddr;
	result = dns_request_createvia(requestmgr, msg, srcaddr, destaddr,
				       0, NULL, FIND_TIMEOUT, global_task,
				       recvns, reqinfo, request);
	check_result(result, "dns_request_createvia");
	requests++;
}

static void
start_reroute(void) {
	isc_result_t result;
	dns_rdataset_t *rdataset = NULL;
	dns_name_t *name = NULL;
	dns_request_t *request = NULL;
	dns_message_t *soaquery = NULL;
	dns_name_t *client_domainname = NULL;
        isc_buffer_t *client_domain_buf = NULL;
        unsigned int len = OUTPUTBUF;
        unsigned int labels = 0;
	
	ddebug("start_reroute()");

	if (userzone != NULL && userserver != NULL) {
		send_reroute(userzone, userserver, localaddr);
		return;
	}

	result = dns_message_create(mctx, DNS_MESSAGE_INTENTRENDER,
				    &soaquery);
	check_result(result, "dns_message_create");

	soaquery->flags |= DNS_MESSAGEFLAG_RD;

	result = dns_message_gettempname(soaquery, &name);
	check_result(result, "dns_message_gettempname");

	result = dns_message_gettemprdataset(soaquery, &rdataset);
	check_result(result, "dns_message_gettemprdataset");

	/* Want to get all authoritative name servers (type NS records) */
	dns_rdataset_makequestion(rdataset, dns_rdataclass_in,
				  dns_rdatatype_ns);

	/* Get domain name of client_hostname */
	result = dns_message_gettempname(soaquery, &client_domainname);
        check_result(result, "dns_message_gettempname");

        result = isc_buffer_allocate(soaquery->mctx, &client_domain_buf, len);
        check_result(result, "isc_buffer_allocate");

	dns_name_setbuffer(client_domainname, client_domain_buf);
        labels = dns_name_countlabels(client_hostname);
        result = dns_name_splitatdepth(client_hostname, labels-1, NULL, client_domainname);
        check_result(result, "dns_name_splitatdepth");

	/* Destroy client_hostname */
	dns_message_puttempname(reroutemsg, &client_hostname);

	/* Put client's domainname into Question section and send */ 
	dns_name_init(name, NULL);
	dns_name_clone(client_domainname, name);
	
	ISC_LIST_INIT(name->list);
	ISC_LIST_APPEND(name->list, rdataset, link);
	dns_message_addname(soaquery, name, DNS_SECTION_QUESTION);

	if (userserver != NULL)
		sendrequest(localaddr, userserver, soaquery, &request);
	else {
		ns_inuse = 0;
		sendrequest(localaddr, &servers[ns_inuse], soaquery, &request);
	}

	isc_buffer_free(&client_domain_buf);
	dns_message_puttempname(soaquery, &client_domainname);
}

static void
cleanup(void) {
	ddebug("cleanup()");

	ddebug("Shutting down task manager");
	isc_taskmgr_destroy(&taskmgr);

	ddebug("Destroying event");
	isc_event_free(&global_event);

	ddebug("Shutting down socket manager");
	isc_socketmgr_destroy(&socketmgr);

	ddebug("Shutting down timer manager");
	isc_timermgr_destroy(&timermgr);

	ddebug("Destroying memory context");
	if (memdebugging)
		isc_mem_stats(mctx, stderr);
	isc_mem_destroy(&mctx);
}

static void
getinput(isc_task_t *task, isc_event_t *event) {
	isc_boolean_t more;

	UNUSED(task);

	if (shuttingdown) {
		maybeshutdown();
		return;
	}

	if (global_event == NULL)
		global_event = event;

	reset_system();
	more = user_interaction();
	if (!more) {
		printf("pid %d will be shut down now\n", getpid());
		isc_app_shutdown();
		return;
	}
	start_reroute();
	return;
}

int
main(int argc, char **argv) {
	isc_result_t result;
	style = &dns_master_style_debug;
	struct timeval tp;
	int rtn;

	input = stdin;

        rtn = gettimeofday(&tp, NULL);             /* get current time */
        t1 = (double)tp.tv_sec+(1.e-6)*tp.tv_usec;

	isc_app_start();

	parse_args(argc, argv);

	setup_system();

	result = isc_app_onrun(mctx, global_task, getinput, NULL);
	check_result(result, "isc_app_onrun");

	starting_pid = getpid();	/* parent of all child processes */
	printf("pid = %d\n", starting_pid);

	(void)isc_app_run();

	cleanup();

	isc_app_finish();

	if (seenerror)
		return (2);
	else
		return (0);
}
