/* openssl_client.c 
 * written by Eric Rescorla,
 * John Viega, Matt Messier, Pravir Chandra,
 * W. Richard Stevens,
 * and David Wilkinson, 09-14-03 
 */

#include <unistd.h>
#include <dns/openssl_client.h>
#include <dns/openssl_common.h>

#include <sys/types.h>
#include "./include/dns/unp.h"

static void
connect_alarm(int signo) {
	return;		/* just interrupt the connect() */
}

int
connect_timeo(int sockfd, const SA *saptr, socklen_t salen, int nsec) {
	Sigfunc *sigfunc;
	int n;

	sigfunc = Signal(SIGALRM, connect_alarm);
	if (alarm(nsec) != 0)
		printf("connect_timeout: alarm was already set\n");

	if ((n = connect(sockfd, saptr, salen)) < 0) {
		if (errno == EINTR)
			errno = ETIMEDOUT;
	}
	alarm(0);			/* turn off the alarm */
	Signal(SIGALRM, sigfunc);	/* restore previous signal handler */

	return(n);
}

SSL_CTX *initialize_client_ctx() {
        SSL_CTX *ctx;

        ctx = initialize_ctx(KEYFILE_CLIENT, PASSWORD_CLIENT);
        SSL_CTX_set_options(ctx, SSL_OP_ALL | SSL_OP_NO_SSLv2);
        if (SSL_CTX_set_cipher_list(ctx, CIPHER_LIST_CLIENT) != 1)
                berr_exit("Error setting cipher list for client (no valid ciphers)");

        return ctx;
}

void ssl_authenticate_server(dns_request_t *request) {
        SSL_CTX *ctx;
        SSL *ssl;
        BIO *sbio;
        int sock;
        int result;
        isc_socket_t *server_socket;
        isc_sockaddr_t address;
	int iteration = 0;

retry:
	/* Get the file descriptor for the current TCP socket */
	server_socket = dns_request_socket(&request);
        address = isc_socket_address(server_socket);
        address.type.sin.sin_port = htons(PORT);
        sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);

	if (sock < 0)
                err_exit("Couldn't create socket");

	sleep(1);	
        printf("pid %d: connecting to %s, ", getpid(), 
	       inet_ntoa(address.type.sin.sin_addr));
        printf("port %d\n", ntohs(address.type.sin.sin_port));

	if (connect_timeo(sock, (struct sockaddr *)&address.type.sin,
			    sizeof(address.type.sin), 2) < 0) {
		close(sock);
		printf("pid %d: Couldn't connect socket\n", getpid());
		exit(0);
	}
		
	RAND_cleanup();

	/* Load randomness */
        if (!(RAND_load_file(RANDOM_CLIENT, 1024)))
                berr_exit("Couldn't load randomness");

        /* Create SSL context */
        ctx = initialize_client_ctx();
	
        /* Connect the SSL socket */
	sbio = BIO_new(BIO_s_socket());
	BIO_set_fd(sbio, sock, BIO_NOCLOSE);
	ssl = SSL_new(ctx);
        SSL_set_bio(ssl, sbio, sbio);
	sleep(1);	/* Let server catch up */
        result = SSL_connect(ssl);

        switch(SSL_get_error(ssl, result))
        {
                case SSL_ERROR_NONE:
                        break;
                case SSL_ERROR_ZERO_RETURN:
                        printf("Zero return error\n");
                        break;
                case SSL_ERROR_WANT_READ:
                        printf("Want read error\n");
                        break;
                case SSL_ERROR_WANT_WRITE:
                        printf("Want write error\n");
                        break;
                default:
                        printf("An SSL_connect error occurred\n");
			iteration++;
			if (iteration < 2) {
			        SSL_shutdown(ssl);
				SSL_free(ssl);
				SSL_CTX_free(ctx);
				close(sock);
				ERR_remove_state(0);
				sleep(3);
				goto retry;
			}
                        break;
        }
	
	if (result <= 0) {
	        printf("pid %d: SSL_connect error, result = %d\n", getpid(), result);
		goto cleanup;
	}
	
	/* Verify authenticity of server's certificate */
	if (SSL_get_verify_result(ssl) != X509_V_OK)
                printf("Certificate doesn't verify\n");
	
cleanup:	
        SSL_shutdown(ssl);
        SSL_free(ssl);
	SSL_CTX_free(ctx);
        close(sock);
	ERR_remove_state(0);
}

